----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    15:08:55 04/04/2011 
-- Design Name: 
-- Module Name:    deser - deser_a 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

library WORK;
use WORK.common.ALL;



entity deser is
   Port (
		input  		: in t_deser_i;
		output 		: out t_deser_o;
		reset			: in std_logic;
		clk40 		: in std_logic
	);
end deser;




architecture deser_a of deser is

	signal i	: t_deser_i;
	signal o : t_deser_o := t_deser_o_init;


	-- After the buffer
	signal data_in				: std_logic;
	signal data_delayed_m 	: std_logic;
	signal data_delayed_s	: std_logic;
	signal data_dly_fabric	: std_logic;
	signal DLY_BUSY 			: std_logic;
	signal DLY_BUSY_M 		: std_logic;
	signal DLY_BUSY_S 		: std_logic;
	
	signal DLY_CAL_M 			: std_logic	:= '0';
	signal DLY_CAL_S 			: std_logic	:= '0';
	signal DLY_RST 			: std_logic	:= '0';
	signal DLY_CE 				: std_logic	:= '0';
	signal DLY_INC 			: std_logic	:= '0';
	signal doing_cals			: std_logic := '0';
	signal do_cals				: std_logic := '0';
	
	signal PD_INCDEC 			: std_logic;
	signal PD_VALID			: std_logic;
	signal PD_EDGE				: std_logic;
	signal ISD_CASCADE		: std_logic;

	constant integrator_bits: integer													:= 16;--6;
	signal integrator_hi		: signed(integrator_bits-1 downto 0)				:= to_signed(1,16);
	signal integrator_lo		: signed(integrator_bits-1 downto 0)				:= to_signed(-1,16);

	
	signal iserdes_q			: std_logic_vector(3 downto 0);
	
	signal s20 					: std_logic := '0';
	
	signal freshword 			: std_logic_vector(19 downto 0) := (others => '0');
	signal ripewords 			: std_logic_vector(39 downto 0) := (others => '0');

	signal iodelay_at_low_limit : std_logic := '0';
	signal iodelay_at_high_limit : std_logic := '0';
	
begin

	i <= input;
	output <= o;

	o.debug(1)  <= data_dly_fabric;
	o.debug(2)  <= DLY_BUSY;
	o.debug(3)  <= DLY_BUSY_M;
	o.debug(4)  <= DLY_BUSY_S;
	o.debug(5)  <= DLY_RST;
	o.debug(6)  <= DLY_CAL_S;
	o.debug(7)  <= DLY_CAL_M;
	o.debug(8)  <= PD_VALID and PD_INCDEC;
	o.debug(9)  <= PD_VALID and not PD_INCDEC;
	o.debug(10) <= DLY_CE and DLY_INC;
	o.debug(11) <= DLY_CE and not DLY_INC ;
	o.debug(12) <= reset;
--	o.debug(13) <= '0';
--	o.debug(14) <= '0';
--	o.debug(15) <= 
--	o.debug(16) <= iodelay_at_low_limit or iodelay_at_high_limit;
--	o.debug(17) <= s20;
--	o.debug(18) <= clk40;
--	o.debug(19) <= s20;
--	o.debug(20) <= clk40;


	-- INPUT BUFFER (IBUFDS)
	ibufds_inst : IBUFDS
		generic map (DIFF_TERM  => TRUE, IOSTANDARD => "LVDS_33")
		port map (I => i.serialin_p, IB => i.serialin_n, O => data_in);
	
   IODELAY2_inst_M : IODELAY2
   generic map (
      COUNTER_WRAPAROUND => "STAY_AT_LIMIT", -- "STAY_AT_LIMIT" or "WRAPAROUND" 
      DATA_RATE => "SDR",                 -- "SDR" or "DDR" 
      DELAY_SRC => "IDATAIN",                  -- "IO", "ODATAIN" or "IDATAIN" 
      IDELAY2_VALUE => 0,                 -- Delay value when IDELAY_MODE="PCI" (0-255)
      IDELAY_TYPE => "DIFF_PHASE_DETECTOR",           -- "FIXED", "DEFAULT", "VARIABLE_FROM_ZERO", "VARIABLE_FROM_HALF_MAX" 
                                          -- or "DIFF_PHASE_DETECTOR" 
      IDELAY_VALUE => 0,                  -- Amount of taps for fixed input delay (0-255)
      SERDES_MODE => "MASTER",              -- "NONE", "MASTER" or "SLAVE" 
      SIM_TAPDELAY_VALUE => 75            -- Per tap delay used for simulation in ps
   )
   port map (
      BUSY => DLY_BUSY_M,         -- 1-bit output: Busy output after CAL
      DATAOUT => data_delayed_m,   -- 1-bit output: Delayed data output to ISERDES/input register
      DATAOUT2 => data_dly_fabric, -- 1-bit output: Delayed data output to general FPGA fabric
      DOUT => open,         -- 1-bit output: Delayed data output
      TOUT => open,         -- 1-bit output: Delayed 3-state output
      CAL => DLY_CAL_M,           -- 1-bit input: Initiate calibration input
      CE => DLY_CE,             -- 1-bit input: Enable INC input
      CLK => i.clk200,           -- 1-bit input: Clock input
      IDATAIN => data_in,   -- 1-bit input: Data input (connect to top-level port or I/O buffer)
      INC => DLY_INC,           -- 1-bit input: Increment / decrement input
      IOCLK0 => i.clk800,     -- 1-bit input: Input from the I/O clock network
      IOCLK1 => '0',     -- 1-bit input: Input from the I/O clock network
      ODATAIN => '0',   -- 1-bit input: Output data input from output register or OSERDES2.
      RST => DLY_RST,           -- 1-bit input: Reset to zero or 1/2 of total delay period
      T => '1'                -- 1-bit input: 3-state input signal
   );

--DLY_BUSY_M <= '0';
--DLY_BUSY_S <= '0';
--data_dly_fabric <= data_in;
--data_delayed <= '0';


   IODELAY2_inst_S : IODELAY2
   generic map (
      COUNTER_WRAPAROUND => "WRAPAROUND", -- "STAY_AT_LIMIT" or "WRAPAROUND" 
      DATA_RATE => "SDR",                 -- "SDR" or "DDR" 
      DELAY_SRC => "IDATAIN",                  -- "IO", "ODATAIN" or "IDATAIN" 
      IDELAY_MODE => "NORMAL",            -- "NORMAL" or "PCI" 
      IDELAY_TYPE => "DIFF_PHASE_DETECTOR",           -- "FIXED", "DEFAULT", "VARIABLE_FROM_ZERO", "VARIABLE_FROM_HALF_MAX" 
                                          -- or "DIFF_PHASE_DETECTOR" 
      IDELAY_VALUE => 0,                  -- Amount of taps for fixed input delay (0-255)
      SERDES_MODE => "SLAVE",              -- "NONE", "MASTER" or "SLAVE" 
      SIM_TAPDELAY_VALUE => 75            -- Per tap delay used for simulation in ps
   )
   port map (
      BUSY => DLY_BUSY_S,         -- 1-bit output: Busy output after CAL
      DATAOUT => data_delayed_s,   -- 1-bit output: Delayed data output to ISERDES/input register
      DATAOUT2 => open, -- 1-bit output: Delayed data output to general FPGA fabric
      DOUT => open,         -- 1-bit output: Delayed data output
      TOUT => open,         -- 1-bit output: Delayed 3-state output
      CAL => DLY_CAL_S,           -- 1-bit input: Initiate calibration input
      CE => DLY_CE,             -- 1-bit input: Enable INC input
      CLK => i.clk200,           -- 1-bit input: Clock input
      IDATAIN => data_in,   -- 1-bit input: Data input (connect to top-level port or I/O buffer)
      INC => DLY_INC,           -- 1-bit input: Increment / decrement input
      IOCLK0 => i.clk800,     -- 1-bit input: Input from the I/O clock network
      IOCLK1 => '0',     -- 1-bit input: Input from the I/O clock network
      ODATAIN => '0',   -- 1-bit input: Output data input from output register or OSERDES2.
      RST => DLY_RST,           -- 1-bit input: Reset to zero or 1/2 of total delay period
      T => '0'                -- 1-bit input: 3-state input signal
   );

DLY_BUSY <= DLY_BUSY_M or DLY_BUSY_S;

--PD_INCDEC <= '0';
--PD_VALID <= '0';
--iserdes_q <= (others => '0');

   ISERDES2_inst_master : ISERDES2
   generic map (
      BITSLIP_ENABLE => FALSE,        -- Enable Bitslip Functionality (TRUE/FALSE)
      DATA_RATE => "SDR",             -- Data-rate ("SDR" or "DDR")
      DATA_WIDTH => 4,                -- Parallel data width selection (2-8)
      INTERFACE_TYPE => "RETIMED", -- "NETWORKING", "NETWORKING_PIPELINED" or "RETIMED" 
      SERDES_MODE => "MASTER"           -- "NONE", "MASTER" or "SLAVE" 
   )
   port map (
      CFB0 => open,           -- 1-bit output: Clock feed-through route output
      CFB1 => open,           -- 1-bit output: Clock feed-through route output
      DFB => open,             -- 1-bit output: Feed-through clock output
      FABRICOUT => open, -- 1-bit output: Unsynchrnonized data output
      INCDEC => PD_INCDEC,       -- 1-bit output: Phase detector output
      -- Q1 - Q4: 1-bit (each) output: Registered outputs to FPGA logic
      Q1 => iserdes_q(0),
      Q2 => iserdes_q(1),
      Q3 => iserdes_q(2),
      Q4 => iserdes_q(3),
      SHIFTOUT => ISD_CASCADE,   -- 1-bit output: Cascade output signal for master/slave I/O
      VALID => PD_VALID,         -- 1-bit output: Output status of the phase detector
      BITSLIP => '0',     -- 1-bit input: Bitslip enable input
      CE0 => '1',             -- 1-bit input: Clock enable input
      CLK0 => i.clk800,           -- 1-bit input: I/O clock network input
      CLK1 => '0',           -- 1-bit input: Secondary I/O clock network input
      CLKDIV => i.clk200,       -- 1-bit input: FPGA logic domain clock input
      D => data_delayed_m,                 -- 1-bit input: Input data
      IOCE => i.serdesstrobe,           -- 1-bit input: Data strobe input
      RST => reset,             -- 1-bit input: Asynchronous reset input
      SHIFTIN => PD_EDGE      -- 1-bit input: Cascade input signal for master/slave I/O
   );

   ISERDES2_inst : ISERDES2
   generic map (
      BITSLIP_ENABLE => FALSE,        -- Enable Bitslip Functionality (TRUE/FALSE)
      DATA_RATE => "SDR",             -- Data-rate ("SDR" or "DDR")
      DATA_WIDTH => 4,                -- Parallel data width selection (2-8)
      INTERFACE_TYPE => "RETIMED", -- "NETWORKING", "NETWORKING_PIPELINED" or "RETIMED" 
      SERDES_MODE => "SLAVE"           -- "NONE", "MASTER" or "SLAVE" 
   )
   port map (
      CFB0 => open,           -- 1-bit output: Clock feed-through route output
      CFB1 => open,           -- 1-bit output: Clock feed-through route output
      DFB => open,             -- 1-bit output: Feed-through clock output
      FABRICOUT => open, -- 1-bit output: Unsynchrnonized data output
      INCDEC => open,       -- 1-bit output: Phase detector output
      -- Q1 - Q4: 1-bit (each) output: Registered outputs to FPGA logic
      Q1 => open,
      Q2 => open,
      Q3 => open,
      Q4 => open,
      SHIFTOUT => PD_EDGE,   -- 1-bit output: Cascade output signal for master/slave I/O
      VALID => OPEN,         -- 1-bit output: Output status of the phase detector
      BITSLIP => '0',     -- 1-bit input: Bitslip enable input
      CE0 => '1',             -- 1-bit input: Clock enable input
      CLK0 => i.clk800,           -- 1-bit input: I/O clock network input
      CLK1 => '0',           -- 1-bit input: Secondary I/O clock network input
      CLKDIV => i.clk200,       -- 1-bit input: FPGA logic domain clock input
      D => data_delayed_s,                 -- 1-bit input: Input data
      IOCE => i.serdesstrobe,           -- 1-bit input: Data strobe input
      RST => reset,             -- 1-bit input: Asynchronous reset input
      SHIFTIN => ISD_CASCADE      -- 1-bit input: Cascade input signal for master/slave I/O
   );



-- State machine that controls the phase detector and issues idelay commands

process (i.clk200)
	variable integrator		: signed(integrator_bits-1 downto 0)				:= (others => '0');
	variable integrator_hiv	: signed(integrator_bits-1 downto 0)				:= to_signed(1,16);
	variable integrator_lov	: signed(integrator_bits-1 downto 0)				:= to_signed(-1,16);
	variable status 			: integer range 7 downto 0 							:= 0;
	variable nextstatus		: integer range 7 downto 0								:= 0;
	constant pulse_enlarge 	: integer 													:= 1;
	variable pulse_enlarger : integer range pulse_enlarge downto 0 			:= pulse_enlarge;
	constant busy_enlarge 	: integer 													:= 1;
	variable busy_enlarger  : integer range busy_enlarge downto 0 				:= busy_enlarge;
	variable current_tap		: unsigned(7 downto 0) := "10000000";
	variable DLY_CE_pipe		: std_logic;
	
begin
	if rising_edge(i.clk200) then
		if reset = '1' then 
			status := 0;
			doing_cals <= '0';
			current_tap := "10000000";
			integrator := (others => '0');
			DLY_CAL_M <= '0';
			DLY_CAL_S <= '0';
			DLY_INC <= '0';
			DLY_CE <= '0';
			DLY_RST <= '0';
		else
			o.reg_int(test1_reg) <= std_logic_vector(current_tap);
			
--			if current_tap <= 128 - unsigned(i.reg(delayrad_reg)) then iodelay_at_low_limit <= '1';
--			else iodelay_at_low_limit <= '0';
--			end if;
--			if current_tap >= 128 + unsigned(i.reg(delayrad_reg)) then iodelay_at_high_limit <= '1';
--			else iodelay_at_high_limit <= '0';
--			end if;
			
			case status is
				when 0 =>
					DLY_CAL_M <= '1';
					DLY_CAL_S <= '1';
					status  := 3;
					nextstatus := 2;
					pulse_enlarger := pulse_enlarge;
					busy_enlarger := busy_enlarge;
				when 1 =>
					DLY_RST <= '1';
					status  := 3;
					nextstatus := 2;
					pulse_enlarger := pulse_enlarge;
					busy_enlarger := busy_enlarge;
				when 2 =>
					doing_cals <= '0';
					
					if do_cals = '1' then 
						doing_cals <= '1';
						DLY_CAL_S <= '1';
						status := 3;
					elsif integrator = integrator_lov then --= 1 then
						DLY_CE <= not iodelay_at_low_limit; -- puesto por si se ejerce un control activo sobre los lmites
						DLY_INC <= '0';
						status := 3;
					elsif integrator = integrator_hiv then --=2**integrator_bits-1 then
						DLY_CE <= not iodelay_at_high_limit; -- puesto por si se ejerce un control activo sobre los lmites
						DLY_INC <= not iodelay_at_high_limit; -- puesto por si se ejerce un control activo sobre los lmites
						status := 3;
					end if;
					nextstatus := 2;
					pulse_enlarger := pulse_enlarge;
					busy_enlarger := busy_enlarge;
					
					if PD_VALID = '1' and PD_INCDEC = '0' then
						integrator := integrator - 1;
					elsif PD_VALID ='1' AND PD_INCDEC = '1' then
						integrator := integrator + 1;
					end if;

				when others =>
					if pulse_enlarger > 0 then pulse_enlarger := pulse_enlarger -1;
					elsif DLY_BUSY = '1' then
						DLY_CAL_M <= '0';
						DLY_CAL_S <= '0';
						DLY_CE <= '0';
						DLY_INC <= '0';
						DLY_RST <= '0';
						integrator := (others => '0');--2**(integrator_bits-1);
					elsif busy_enlarger > 0 then busy_enlarger := busy_enlarger-1;
					else status := nextstatus;
					end if;
			end case;
			
			if 	DLY_CE = '1' and DLY_CE_pipe = '0' and DLY_INC = '0' then --and current_tap > 0 then 
				current_tap := current_tap - 1;
			elsif DLY_CE = '1' and DLY_CE_pipe = '0' and DLY_INC = '1' then--and current_tap < 255 then 
				current_tap := current_tap + 1;
			end if;
			DLY_CE_pipe := DLY_CE;
			
			integrator_hiv := integrator_hi;
         integrator_lov := integrator_lo;

		end if;
	end if;
end process;

o.reg_ch_int(test1_reg) <= s20; -- el cambio de los datos se hace a clk200, pero la orden va con clk40


-- Build the 20-bit words for output
process (i.clk200)
	variable s20old 	: std_logic 							:= '0';
	variable cycle 	: integer range 4 downto 0			:= 0;
	variable nextword : std_logic_vector(19 downto 0)	:= (others => '0');
begin
	if rising_edge(i.clk200) then
		nextword(cycle*4+3 downto cycle*4) := iserdes_q(3 downto 0);
		if s20old /= s20 then 
			freshword <= nextword;
			cycle := 0;
		else cycle := cycle + 1;
		end if;
		s20old := s20;
	end if;
end process;


process (clk40)
	variable offset 			: integer range 19 downto 0;
	variable goodword 		: std_logic_vector(19 downto 0);
	variable startstopfail	: std_logic											:= '0';
	constant relock_times	: integer											:= 100;
	variable relock 			: integer range relock_times downto 0		:= relock_times;
--	constant unlocking_times: integer											:= 3;
	variable unlocking		: unsigned(7 downto 0)							:= unsigned(reg_init(tsc_unlocking));
	
	constant recal_timer_bits: integer													:= 32;
	variable recal_time		: unsigned(recal_timer_bits - 1 downto 0)			:= (others => '1');
	variable recal_timer		: unsigned(recal_timer_bits - 1 downto 0)			:= (others => '1');

begin
	if rising_edge(clk40) then
		if reset = '1' then 
			offset := 0;
			relock := relock_times;
			o.lock <= '0';
			
			-- RECALIBRATION TIMER (NOTHING TO DO WITH ALIGNMENT)
			recal_time := to_unsigned(2**(to_integer(unsigned(i.reg(log2_recaltime_reg)(5 downto 0)))+1)-1,recal_timer_bits);
			recal_timer := recal_time;
		else
			s20 <= not s20;
			-- RECALIBRATION TIMER (NOTHING TO DO WITH ALIGNMENT)
			if recal_timer = 0 then do_cals <= '1';--(recal_timer(recal_timer_bits-1 downto 5) & recal_timer(4 downto 0)) = 0 then do_recal := '1';--esto es para hacer rfagas de cal
			elsif doing_cals = '1' then do_cals <= '0';
			end if;
			if recal_timer > 0 then recal_timer := recal_timer - 1;
			else recal_timer := recal_time;
			end if;
			recal_time := to_unsigned(2**(to_integer(unsigned(i.reg(log2_recaltime_reg)(5 downto 0)))+1)-1,recal_timer_bits);
			-- INTEGRATOR LIMITIS (NOTHING TO DO WITH ALIGNMENT)
			integrator_hi <= to_signed(2**(to_integer(unsigned(i.reg(log2_integr_reg)(3 downto 0)))+1)-1,integrator_bits);
			integrator_lo <= to_signed(-2**(to_integer(unsigned(i.reg(log2_integr_reg)(3 downto 0)))+1)+1,integrator_bits);
			
			
			ripewords(39 downto 20) <= freshword;
			ripewords(19 downto 0) <= ripewords(39 downto 20);
			
			for ind in 19 downto 0 loop
				goodword(ind) := ripewords(19 - ind + offset);
			end loop;
			o.word <= goodword;
			
			if goodword(0) /= '1' or goodword(19) /= '0' then startstopfail := '1';
			else startstopfail := '0';
			end if;
			
			if o.lock = '0' and startstopfail = '1' then --estoy unlock y viene una mala
				relock := relock_times;
				if offset = 19 then offset := 0;
				else offset := offset + 1;
				end if;
			elsif o.lock = '0' and relock > 0 then  -- sigo unlock, viene una buena pero todava no estoy completamente relock
				relock := relock - 1;
			elsif o.lock = '0' then --segua unlock y llega una buena palabra y estaba a punto de relock
				o.lock <= '1';
				unlocking := unsigned(i.reg(tsc_unlocking));
			elsif startstopfail = '1' and unlocking > 0 then --estoy en lock y llega una mala palabra, pero todava no estoy completamente unlock
				unlocking := unlocking - 1;
			elsif startstopfail = '1' then --estaba en lock, pero llega una nueva mala y ya estaba al lmite
				o.lock <= '0';
				relock := relock_times;
			elsif unlocking < unsigned(i.reg(tsc_unlocking)) then --estaba en lock, llega una buena, y haba empezado a unlock
				unlocking := unlocking + 1;
			end if;
			
			o.iodelay_at_limit <= iodelay_at_low_limit or iodelay_at_high_limit;
		end if;
	end if;
end process;

end deser_a;

