----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    14:04:26 07/06/2012 
-- Design Name: 
-- Module Name:    ofcu_ros - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

library UNISIM;
use UNISIM.VComponents.all;

library WORK;
use WORK.common.ALL;

entity ofcu_ros is
	Port (
		sc_bsel_n	: in std_logic;
		sc_wr_n 		: in std_logic;
		sc_data 		: inout std_logic_vector(7 downto 0) := (others => 'Z');
		sc_add	 	: in std_logic_vector(3 downto 0);
		         	
		sd_tsc		: in std_logic;
		sd_ch24		: inout std_logic;
		sd				: inout std_logic_vector(2 downto 1);
		rx_en		 	: out std_logic_vector(2 downto 1);
		sq_en			: out std_logic_vector(2 downto 1);
		en_sd			: out std_logic_vector(2 downto 1);
		
		trg_data		: inout	std_logic_vector(15 downto 0);
		trg_data_dir: out std_logic;
      trg_strobe	: inout std_logic;
      ros_ready	: inout std_logic;
		evcntres		: in std_logic;
		l1a			: in std_logic;
      serialin_p 	: in std_logic;
      serialin_n	: in std_logic;
		
		leds			: out std_logic_vector(5 downto 0);
		
		onewire1		: inout std_logic;
		onewire3		: inout std_logic;
		
		Dir_spare	: in	std_logic;
		BSpare		: in std_logic_vector(4 downto 0);
		B_SP_indiv	: out std_logic;
		read_mode	: in std_logic;
		
		oftx_p 		: out std_logic_vector(2 downto 1);
		oftx_n 		: out std_logic_vector(2 downto 1);
	   oftx_dis		: out std_logic_vector(2 downto 1);
		
		testconn	 	: inout std_logic_vector(20 downto 1);
		
		ixo_on		: in STD_LOGIC;
		clk40pad 	: in  STD_LOGIC;
		breset 		: in  STD_LOGIC);
end ofcu_ros;

architecture Behavioral of ofcu_ros is
	signal to_registers : t_registers_i;
	signal from_registers : t_registers_o;

	signal to_onewire1 : t_onewire_i;
	signal from_onewire1 : t_onewire_o;

	signal to_onewire3 : t_onewire_i;
	signal from_onewire3 : t_onewire_o;

	signal to_receivers : t_receivers_i;
	signal from_receivers : t_receivers_o;

	signal to_serialclk : t_serialclk_i;
	signal from_serialclk : t_serialclk_o;

	signal to_deser : t_deser_i;
	signal from_deser : t_deser_o;
	signal long_sw_rst_serdes	: std_logic;

	signal to_rosj3 : t_rosj3_i;
	signal from_rosj3 : t_rosj3_o;

	signal to_serializer : t_serializer_i;
	signal from_serializer : t_serializer_o;
	signal oftx : std_logic_vector(2 downto 1);

	signal to_tscj3 : t_tscj3_i;
	signal from_tscj3 : t_tscj3_o;

	signal to_statusleds : t_statusleds_i;
	signal from_statusleds : t_statusleds_o;

	signal to_test : t_test_i;
	signal from_test : t_test_o;

	signal clk40 : std_logic;

begin

	

   STARTUP_SPARTAN6_inst : STARTUP_SPARTAN6
   port map (
      CFGCLK => open,       -- 1-bit output: Configuration logic main clock output.
      CFGMCLK => open,     -- 1-bit output: Configuration internal oscillator clock output.
      EOS => open,             -- 1-bit output: Active high output signal indicates the End Of Configuration.
      CLK => '0',             -- 1-bit input: User startup-clock input
      GSR => breset,             -- 1-bit input: Global Set/Reset input (GSR cannot be used for the port name)
      GTS => '0',             -- 1-bit input: Global 3-state input (GTS cannot be used for the port name)
      KEYCLEARB => '1'  -- 1-bit input: Clear AES Decrypter Key input from Battery-Backed RAM (BBRAM)
   );

	





--********************************************************************************
--	<REGISTERS>
--********************************************************************************
	inst_registers : registers
	port map(
		input  => to_registers,
		output => from_registers,
		sc_data => sc_data,
		clk40 => clk40
	);
	
	to_registers.sc_add		<= sc_add	;
	to_registers.sc_bsel		<= not sc_bsel_n	;
	to_registers.sc_wr 		<= not sc_wr_n		;
	
--********************************************************************************
--	</REGISTERS>
--********************************************************************************
	
	
	
--********************************************************************************
--	<ONE WIRE>
--********************************************************************************

	ow_gen: if not exclude1wire generate
	
		inst_onewire1 : onewire
		generic map(
			ctrl_reg => ow1_ctrl_reg,
			read_reg => ow1_read_reg,
			crc_reg 	=> ow1_crc_reg,
			write_reg => ow1_write_reg)
		port map(
			input  => to_onewire1,
			output => from_onewire1,
			wire => onewire1,
			clk40 => clk40	);

		to_onewire1.reg 			<= from_registers.reg;
		to_onewire1.ext_wr 		<= from_registers.ext_wr;
		to_onewire1.reg_ch_int 	<= from_registers.reg_ch_int;
		
		to_registers.reg_int(ow1_read_reg) 			<= from_onewire1.reg_int(ow1_read_reg);
		to_registers.reg_int(ow1_crc_reg) 			<= from_onewire1.reg_int(ow1_crc_reg);
		to_registers.reg_ch_int(ow1_read_reg) 		<= from_onewire1.reg_ch_int(ow1_read_reg);
		to_registers.reg_ch_int(ow1_crc_reg) 		<= from_onewire1.reg_ch_int(ow1_crc_reg);
			
		inst_onewire3 : onewire
		generic map(
			ctrl_reg => ow3_ctrl_reg,
			read_reg => ow3_read_reg,
			crc_reg 	=> ow3_crc_reg,
			write_reg => ow3_write_reg)
		port map(
			input  => to_onewire3,
			output => from_onewire3,
			wire => onewire3,
			clk40 => clk40	);
			
		to_onewire3.reg 			<= from_registers.reg;
		to_onewire3.ext_wr 		<= from_registers.ext_wr;
		to_onewire3.reg_ch_int 	<= from_registers.reg_ch_int;
		
		to_registers.reg_int(ow3_read_reg) 			<= from_onewire3.reg_int(ow3_read_reg);
		to_registers.reg_int(ow3_crc_reg) 			<= from_onewire3.reg_int(ow3_crc_reg);
		to_registers.reg_ch_int(ow3_read_reg) 		<= from_onewire3.reg_ch_int(ow3_read_reg);
		to_registers.reg_ch_int(ow3_crc_reg) 		<= from_onewire3.reg_ch_int(ow3_crc_reg);

	end generate;

	
	not_ow_gen: if exclude1wire generate
		onewire1 <= 'Z';
		onewire3 <= 'Z';
	end generate;

--********************************************************************************
--	</ONE WIRE>
--********************************************************************************

--********************************************************************************
--	<RECEIVERS>
--********************************************************************************
	inst_receivers : receivers
	port map(
		input  => to_receivers,
		output => from_receivers,
		clk40 => clk40	);
	
	to_receivers.reg 			<= from_registers.reg;
	to_receivers.ext_wr 		<= from_registers.ext_wr;
	to_receivers.reg_ch_int <= from_registers.reg_ch_int;
	to_receivers.sd_tsc 		<= sd_tsc;
	to_receivers.sd_ch24 	<= sd_ch24;
	to_receivers.sd 			<= sd;
	
	to_registers.reg_int(sd_status_reg) 	<= from_receivers.reg_int(sd_status_reg);
	to_registers.reg_ch_int(sd_status_reg) <= from_receivers.reg_ch_int(sd_status_reg);
   oftx_dis						<= from_receivers.tx_dis;
	rx_en 						<= from_receivers.rx_en;
	sq_en 						<= from_receivers.sq_en;
	en_sd 						<= from_receivers.en_sd;
	
	receivers_ofcuros_gen: if board_type = ofcuros generate
		sd_ch24 <= 'Z';
		sd <= "ZZ";
	end generate;

	receivers_tscrear_gen: if board_type = tscrear generate
		sd_ch24 <= from_receivers.sq_en(1);
		sd <= from_receivers.en_sd;
	end generate;

	
--********************************************************************************
--	</RECEIVERS>
--********************************************************************************

--********************************************************************************
--	<SERIALCLK>
--********************************************************************************

	inst_serialclk : serialclk
	port map(
		input  => to_serialclk,
		output => from_serialclk,
		reset => breset or from_registers.ext_wr(ow1_crc_reg)
		);

		to_serialclk.clk40pad		<= clk40pad;
		to_serialclk.reg 				<= from_registers.reg;
	
		clk40 							<= from_serialclk.clk40;

--********************************************************************************
--	</SERIALCLK>
--********************************************************************************

--********************************************************************************
--	<DESER>
--********************************************************************************
	deser_ofcuros_gen: if board_type = ofcuros generate

		inst_deser : deser
		port map(
			input  => to_deser,
			output => from_deser,
			reset => breset or not from_serialclk.bufpll_locked(serialin_bank) or long_sw_rst_serdes,
			clk40 => from_serialclk.clk40s	
			);

		to_deser.serialin_p 		<= serialin_p;
		to_deser.serialin_n 		<= serialin_n;
		to_deser.clk200 			<= from_serialclk.clk200;
		to_deser.clk800 			<= from_serialclk.clk800(serialin_bank);
		to_deser.serdesstrobe 	<= from_serialclk.serdesstrobe(serialin_bank);
		to_deser.bufpll_locked 	<= from_serialclk.bufpll_locked(serialin_bank);
		to_deser.reg 				<= from_registers.reg;
		to_deser.reg_ch_int		<= from_registers.reg_ch_int;

		to_registers.reg_ch_int(test1_reg) 	<= from_deser.reg_ch_int(test1_reg) 	;
		to_registers.reg_int(test1_reg) 		<= from_deser.reg_int(test1_reg) 	;
--		testconn(18 downto 1) 					<= from_deser.debug(18 downto 1);
		
		process(clk40)
			variable tmp_ext_wr : std_logic := '0';
			variable fresh : unsigned(fresh_bits - 2 downto 0) := (others => '1');
		begin
			if rising_edge(clk40) then
				long_sw_rst_serdes <= from_registers.ext_wr(tsc_status_2_la) or tmp_ext_wr;
				tmp_ext_wr := from_registers.ext_wr(tsc_status_2_la);
				if fresh > 0 then
					fresh := fresh - 1;
					if fresh < 4 then long_sw_rst_serdes <= '1';
					end if;
				end if;
			end if;
		end process;
	end generate;
	
	
	serialin_gen: if board_type = tscrear generate
		ibufds_inst : IBUFDS
			generic map (DIFF_TERM  => TRUE, IOSTANDARD => "LVDS_33")
			port map (I => serialin_p, IB => serialin_n, O => open);
	end generate;

--********************************************************************************
--	</DESER>
--********************************************************************************

--********************************************************************************
--	<ROSJ3>
--********************************************************************************
	rosj3_ofcuros_gen: if board_type = ofcuros generate

		trg_data_dir <= '1'; -- 0: TSC_rear, 1: OFCu
		ros_ready <= 'Z';

		inst_rosj3 : rosj3
		port map(
			input  => to_rosj3,
			output => from_rosj3,
			clk40 => clk40	);

		to_rosj3.ros_ready			<= ros_ready;

		process (clk40)
		begin
			if falling_edge(clk40) then
				to_rosj3.evcntres				<= evcntres;
				to_rosj3.l1a 					<= l1a;
			end if;
		end process;

		to_rosj3.reg					<= from_registers.reg;
		to_rosj3.reg_ch_int			<= from_registers.reg_ch_int;
		to_rosj3.ext_wr				<= from_registers.ext_wr;
		
		to_rosj3.bufpll_locked 		<= from_serialclk.bufpll_locked(serialin_bank);
		to_rosj3.pll_locked 			<= from_serialclk.pll_locked;
		to_rosj3.iodelay_at_limit 	<= from_deser.iodelay_at_limit;
		to_rosj3.deser_lock 			<= from_deser.lock;
		to_rosj3.ixo_on 				<= ixo_on;
		
		to_rosj3.word 					<= from_deser.word;
		
		to_rosj3.clk40s			 	<= from_serialclk.clk40s;

		
		trg_data 						<= from_rosj3.trg_data;
		trg_strobe 						<= from_rosj3.trg_strobe;

		testconn(18 downto 1) 					<= from_rosj3.debug(18 downto 1);
		
		to_registers.reg_ch_int(tsc_status_0_la) 	<= from_rosj3.reg_ch_int(tsc_status_0_la) 	;
		to_registers.reg_ch_int(tsc_status_1_la) 	<= from_rosj3.reg_ch_int(tsc_status_1_la) 	;
		to_registers.reg_ch_int(tsc_status_2_la) 	<= from_rosj3.reg_ch_int(tsc_status_2_la) 	;
		to_registers.reg_int(tsc_status_0_la) 		<= from_rosj3.reg_int(tsc_status_0_la) 	;
		to_registers.reg_int(tsc_status_1_la) 		<= from_rosj3.reg_int(tsc_status_1_la) 	;
		to_registers.reg_int(tsc_status_2_la) 		<= from_rosj3.reg_int(tsc_status_2_la) 	;
		
		to_registers.reg_ch_int(test0_reg) 	<= from_rosj3.reg_ch_int(test0_reg) 	;
		to_registers.reg_int(test0_reg) 		<= from_rosj3.reg_int(test0_reg) 	;
		

	end generate;
--********************************************************************************
--	</ROSJ3>
--********************************************************************************

--********************************************************************************
--	<TSCJ3>
--********************************************************************************
	tscj3_tscrear_gen: if board_type = tscrear generate

		trg_data_dir <= '0';  -- 0: TSC_rear, 1: OFCu
		trg_strobe <= 'Z';

		inst_tscj3 : tscj3
		port map(
			input  => to_tscj3,
			output => from_tscj3,
			clk40 => clk40	);
		
		to_tscj3.trg_data <= trg_data;
		to_tscj3.trg_strobe <= trg_strobe;

		process (clk40)
		begin
			if falling_edge(clk40) then
				to_tscj3.evcntres				<= evcntres;
				to_tscj3.l1a 					<= l1a;
			end if;
		end process;

		to_tscj3.bufpll_locked 		<= from_serialclk.bufpll_locked(serialout_bank);
		to_tscj3.pll_locked 			<= from_serialclk.pll_locked;
		to_tscj3.clk40s 				<= from_serialclk.clk40s;
		to_tscj3.ixo_on 				<= ixo_on;

		to_tscj3.reg					<= from_registers.reg;
		to_tscj3.reg_ch_int			<= from_registers.reg_ch_int;
		to_tscj3.ext_wr 				<= from_registers.ext_wr;
		
		ros_ready 						<= from_tscj3.ros_ready;
		testconn(18 downto 1) 		<= from_tscj3.debug(18 downto 1);
		
		to_registers.reg_ch_int(tsc_status_0_la) 	<= from_tscj3.reg_ch_int(tsc_status_0_la) 	;
		to_registers.reg_ch_int(tsc_status_1_la) 	<= from_tscj3.reg_ch_int(tsc_status_1_la) 	;
		to_registers.reg_ch_int(tsc_status_2_la) 	<= from_tscj3.reg_ch_int(tsc_status_2_la) 	;
		to_registers.reg_int(tsc_status_0_la) 		<= from_tscj3.reg_int(tsc_status_0_la) 	;
		to_registers.reg_int(tsc_status_1_la) 		<= from_tscj3.reg_int(tsc_status_1_la) 	;
		to_registers.reg_int(tsc_status_2_la) 		<= from_tscj3.reg_int(tsc_status_2_la) 	;
	end generate;
	
--********************************************************************************
--	</TSCJ3>
--********************************************************************************


--********************************************************************************
--	<SERIALIZER>
--********************************************************************************
	serializer_gen: if board_type = tscrear generate
	
		inst_serializer : serializer
		port map(
			input  => to_serializer,
			output => from_serializer,
			reset => breset or not from_serialclk.bufpll_locked(serialout_bank) or long_sw_rst_serdes,
			clk40 => from_serialclk.clk40s	);
		
		to_serializer.word			<= from_tscj3.word when from_registers.reg(ow1_ctrl_reg)(7 downto 6) = "00" else from_test.word;
		to_serializer.clk200     	<= from_serialclk.clk200;
		to_serializer.clk800    	<= from_serialclk.clk800(serialout_bank);
		to_serializer.serdesstrobe	<= from_serialclk.serdesstrobe(serialout_bank);
		to_serializer.reg				<= from_registers.reg;

		oftx(2) <= from_serializer.serialout;
--		testconn(18 downto 1) 		<= from_serializer.debug(18 downto 1);


		process(clk40)
			variable tmp_ext_wr : std_logic := '0';
		begin
			if rising_edge(clk40) then
				long_sw_rst_serdes <= from_registers.ext_wr(tsc_status_2_la) or tmp_ext_wr;
				tmp_ext_wr := from_registers.ext_wr(tsc_status_2_la);
			end if;
		end process;
	end generate;

	clkout_gen: if board_type = ofcuros generate
		ODDR2_inst : ODDR2
		port map (Q => oftx(2),C0 => clk40,C1 => not clk40,CE => '1',D0 => '1',D1 => '0',R => '0',S => '0');
	end generate;

	oftx(1) <= '0';
	
	inst_obufds : OBUFDS
		generic map (IOSTANDARD => "LVDS_33")
		port map (O => oftx_p(1), OB => oftx_n(1), I => oftx(1));

	inst_obufds2 : OBUFDS
		generic map (IOSTANDARD => "LVDS_33")
		port map (O => oftx_p(2), OB => oftx_n(2), I => oftx(2));


	
--********************************************************************************
--	</SERIALIZER>
--********************************************************************************


--********************************************************************************
--	<STATUSLEDS>
--********************************************************************************
	inst_statusleds : statusleds
	port map(
		input  => to_statusleds,
		output => from_statusleds,
		clk40 => clk40	);
	
	to_statusleds.deser_lock	<= from_deser.lock;
	to_statusleds.sc_bsel_n 	<= sc_bsel_n;
	to_statusleds.evcntres		<= evcntres;
	to_statusleds.l1a				<= l1a;	
	to_statusleds.reg				<= from_registers.reg;
	
	leds(3 downto 0) <= not from_statusleds.leds(3 downto 0);
	
	
--********************************************************************************
--	</STATUSLEDS>
--********************************************************************************

--********************************************************************************
--	<TEST>
--********************************************************************************
	inst_test : test
	port map(
		input  => to_test,
		output => from_test,
		testconn => open,
		reset => breset,
		clk40 => clk40	);
	
	to_test.ext_wr 		<= from_registers.ext_wr;
	to_test.reg_ch_int 	<= from_registers.reg_ch_int;
	to_test.reg				<= from_registers.reg;
	
	to_test.sc_bsel		<= not sc_bsel_n;
	to_test.read_mode		<= read_mode;
	to_test.BSpare 		<= BSpare;
	
	to_test.deser_lock 	<= from_deser.lock;
	to_test.word		 	<= from_deser.word;
	
	to_test.clk40s	 	<= from_serialclk.clk40s;
--	not_invert_word2: for i in 19 downto 0 generate to_test.word(i) <= from_deser.word(i);
--	end generate;

	
	to_registers.reg_int(test2_reg) 		<= from_test.reg_int(test2_reg);
	to_registers.reg_ch_int(test2_reg) 	<= from_test.reg_ch_int(test2_reg);
	
	b_sp_indiv <= '0'; --testconn(1); -- LINE not used, arbitrary value.
	
	-- hidden leds
	leds(4) <= not from_test.lock;
	leds(5) <= '1';
		
	ODDR2_inst : ODDR2
	port map (Q => testconn(20),C0 => clk40,C1 => not clk40,CE => '1',D0 => '1',D1 => '0',R => '0',S => '0');
	
	ODDR2_inst2 : ODDR2
	port map (Q => testconn(19),C0 => from_serialclk.clk40s,C1 => not from_serialclk.clk40s, CE => '1',D0 => '1',D1 => '0',R => '0',S => '0');
	
--********************************************************************************
--	</TEST>
--********************************************************************************




end Behavioral;

